/*
 * $Id: dummies.h,v 1.9 2001/05/14 18:04:39 lindahl Exp $
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.0
 * 
 * Copyright (c) 1991-2001
 * BIOSON Research Institute, Dept. of Biophysical Chemistry
 * University of Groningen, The Netherlands
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * Do check out http://www.gromacs.org , or mail us at gromacs@gromacs.org .
 * 
 * And Hey:
 * Good ROcking Metal Altar for Chronical Sinners
 */

#ifndef _dummies_h
#define _dummies_h

static char *SRCID_dummies_h = "$Id: dummies.h,v 1.9 2001/05/14 18:04:39 lindahl Exp $";
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include "typedefs.h"

typedef struct {
  int nprevdum; /* how many dummy particles are nonlocal */     
  int nnextdum;
  int *idxprevdum; /* index of nonlocal dummy particles */
  int *idxnextdum;
  int nprevconstr; /* how many constr. atoms are nonlocal */
  int nnextconstr;
  int *idxprevconstr; /* indices of nonlocal constructing atoms */
  int *idxnextconstr;
} t_comm_dummies;

/* Communication routines for dummies. The coordinates and
 * forces are only move on a need-to-know basis, usually only
 * 2-3 atoms per processor. To achieve this small amount of
 * communication, and to limit it to nearest neighbour messages,
 * we demand that dummies are not spread over nonadjacent nodes.
 * Thus, keep your dummies close to your constructing atoms.
 * (mdrun & grompp will report an error otherwise)
 */ 

extern void move_construct_x(t_comm_dummies *dummycomm,rvec x[],t_commrec *cr); 
/* Move coords of nonlocal constructing atoms */

extern void move_dummy_xv(t_comm_dummies *dummycomm,rvec x[],rvec v[],t_commrec *cr);
/* Send the coordinates and velocity of a constructed dummy to the home node */

extern void move_dummy_f(t_comm_dummies *dummycomm,rvec f[],t_commrec *cr);
/* Get dummy forces from the home node */

extern void move_construct_f(t_comm_dummies *dummycomm,rvec f[],t_commrec *cr);
/* Send spreaded forces to nonlocal constructing atoms */


extern void construct_dummies(FILE *log,rvec x[],t_nrnb *nrnb,
			      real dt,rvec v[],t_idef *idef);
/* Create positions of dummy atoms based on surrounding atoms.
 */
 
extern void spread_dummy_f(FILE *log,rvec x[],rvec f[],rvec buf[],
			   t_nrnb *nrnb,t_idef *idef);
/* Spread the force operating on the dummy atoms on the surrounding atoms.
 */

#endif

